C  ---------------------------------------------------------------------
      SUBROUTINE OPTIMA ( N, M, XP, XV, XLL, XRL, IXAT, ICODE, FINP,
     &                    FOUT, GRMS, NF, NG, NH, NJ )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is an interface routine. It interfaces a user's program with
C    the Merlin optimization package. The user is expected to supply
C    an input file (FINP) containing either a sequence of Merlin
C    commands, or the object code of an MCL program.
C    Optionaly the user may supply an initial point (XP), lower (XLL),
C    and upper (XRL) bounds, and fix statuses (IXAT) for the parameters.
C    The routine prepares a macro file in order to perform the desired
C    initialization and modifies the Merlin configuration file (CONFIG)
C    so that Merlin reads its input from the supplied input file (FINP).
C    Upon completion of the routine, the minimizer is returned, all
C    temporary files are deleted and the Merlin configuration file is
C    restored to its original contents.
C
C  Input arguments:
C    N         Number of parameters of the objective function.
C    M         Number of terms in case the objective function is a
C              sum of squared terms; zero otherwise.
C    FINP      The file name where the instructions for Merlin reside.
C              The commands should be written as if they were entered
C              from the keyboard.
C    FOUT      The file name where Merlin output will be disposed.
C              If FOUT is blank (ie FINP=' ') the standard output
C              device is used.
C              On a unix system use '/dev/null' to trash the output.
C    ICODE(4)  Integer array. Each element can be either 0 or 1.
C                If ICODE(1)=1 then array XP(N) is used as an initial
C                point; otherwise no initialization occurs.
C
C                If ICODE(2)=1 then array XLL(N) is used to set the
C                lower parameter bounds; otherwise no lower bounds
C                are set.
C                Note that if any element of array XLL is set to a
C                value less than -BLIM, the corresponding lower bound
C                is not set.
C
C                If ICODE(3)=1 then array XRL(N) is used to set the
C                upper parameter bounds otherwise no upper bounds
C                are set.
C                Note that if any element of array XLL is set to a
C                value greater than BLIM, the corresponding upper bound
C                is not set.
C
C                If ICODE(4)=1 then array IXAT(N) is used to set the
C                fix statuses of the parameters else no fix statuses
C                are set.
C
C    XLL(N)    Left margins (lower bounds).
C    XRL(N)    Right margins (upper bounds).
C    IXAT(N)   Fix statuses. 0 for fixed, 1 for a free parameter.
C
C  Output arguments:
C    XV     The value of the objective function at the minimizer.
C    GRMS   The Root Mean Square Gradient.
C    NF     Number objective function evaluations performed by Merlin.
C    NG     Number of calls to the user supplied routine GRANAL that
C           were performed by Merlin.
C    NH     Number of calls to the user supplied routine HANAL that
C           were performed by Merlin.
C    NJ     Number of calls to the user supplied routine JANAL that
C           were performed by Merlin.
C
C  Input / Output arguments:
C    XP(N)  On input if ICODE(1)=1, XP holds the initial point.
C           On output the point where the minimum occurs.
C
C  Notes:
C    1) The following Merlin routines are used:
C       NUNIT, LENGTH, POSIT, GETX, GETVAL, GETCNT, I2STR.
C
C    2) Since SUBROUTINE OPTIMA is called before any initializations
C       are performed by Merlin, care must be taken to ensure that the
C       above routines operate correctly.
C
C    3) This routine aborts (STOPs) on all errors.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
C  Name of the Merlin configuration file.
      CHARACTER*(*) CFILE
      PARAMETER ( CFILE = 'CONFIG' )
C
C  Name of the temporary Merlin configuration file.
      CHARACTER*(*) CFTMP
      PARAMETER ( CFTMP = 'CONFIG.OLD' )
C
C  Name of the temporary macro file.
      CHARACTER*(*) MACRO
      PARAMETER ( MACRO = 'MACTEMP' )
C
C  Large number.
      PARAMETER ( BLIM = 1.E+20 )
C
C  Maximum number of parameters.
      PARAMETER ( MXV = $(MXV) )
C
C  Maximum number of terms.
      PARAMETER ( MXT = $(MXT) )
C
C  Size of the Merlin workspace.
      PARAMETER ( MAXW = $(MAXW) )
C  ---------------------------------------------------------------------
C  Arguments:
      DIMENSION XP(*), XLL(*), XRL(*), IXAT(*), ICODE(4)
      CHARACTER*(*) FINP, FOUT
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /UALLOC/ MINU, MAXU, LASTU
C  ---------------------------------------------------------------------
C  Local variables:
      DIMENSION VERSIM(MAXW)
      PARAMETER ( LCLEN = 80 )
      CHARACTER*(LCLEN) LINE
      CHARACTER*3 MCLCHA
      CHARACTER*10 ICHA
      DIMENSION GRAD(MXV)
      LOGICAL INEXI, CFGEXI, MCLIN
C
C  Check whether the input arguments make any sense.
      IF (N.LT.1) THEN
         WRITE (*,*) 'OPTIMA: The number of parameters ',
     &               'should be at least 1.'
         STOP
      END IF
      IF (N.GT.MXV) THEN
         WRITE (*,*) 'OPTIMA: The number of parameters ',
     &               'should be no more than ', MXV
         STOP
      END IF
      IF (M.LT.0) THEN
         WRITE (*,*) 'OPTIMA: The number of terms ',
     &               'should be at least 0.'
         STOP
      END IF
      IF (M.GT.MXT) THEN
         WRITE (*,*) 'OPTIMA: The number of terms ',
     &               'should be no more than ', MXT
         STOP
      END IF
      LEFIN = LENGTH(FINP)
      IF (LEFIN.EQ.0) THEN
         WRITE (*,*) 'OPTIMA: The input file name is blank.'
         STOP
      END IF
      LEFOUT = LENGTH(FOUT)
C
C  Make sure the input file exists.
      INQUIRE (FILE=FINP,EXIST=INEXI,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         WRITE (*,*) 'OPTIMA: Cannot inquire the input file ',
     &               FINP(1:LEFIN)
         WRITE (*,*) '        Error code: ',IOE
         STOP
      ELSE
         IF (.NOT.INEXI) THEN
            WRITE (*,*) 'OPTIMA: The input file ',FINP(1:LEFIN),
     &                  ' does not exist.'
            STOP
         END IF
      END IF
C
C  Initial values for the unit allocator.
      MINU = 50
      MAXU = 80
      LASTU = MINU-1
C
C  Allocate the necessary unit numbers.
      IU = NUNIT()
      IUC = NUNIT()
      IF (IU.EQ.-1 .OR. IUC.EQ.-1) THEN
         WRITE (*,*) 'OPTIMA: Cannot allocate a free unit.'
         STOP
      END IF
C
C  Check whether the input is an MCL object code.
      OPEN (IU,FILE=FINP)
      READ (IU,'(A)',END=100) MCLCHA
      CLOSE (IU)
      MCLIN = MCLCHA.EQ.'MCL'
C
C  Create the temporary macro file.
      OPEN (IU,FILE=MACRO)
      WRITE (IU,130) 'NOPRINT'
C
C  Check if an initial lower bound (left margin) is required.
C  In that case write the contents of array XLL(1:N) to a file.
      IF (ICODE(2).EQ.1) THEN
         DO 2,I=1,N
            IF (XLL(I).GT.-BLIM .AND. XLL(I).LT.BLIM) THEN
               CALL I2STR(I,ICHA,LECHA)
               WRITE (IU,110) 'LMARGIN', ICHA(1:LECHA), XLL(I)
            END IF
2        CONTINUE
      END IF
C
C  Check if an initial upper bound (right margin) is required.
C  In that case write the contents of array XRL(1:N) to a file.
      IF (ICODE(3).EQ.1) THEN
         DO 3,I=1,N
            IF (XRL(I).GT.-BLIM .AND. XRL(I).LT.BLIM) THEN
               CALL I2STR(I,ICHA,LECHA)
               WRITE (IU,110) 'RMARGIN', ICHA(1:LECHA), XRL(I)
            END IF
3        CONTINUE
      END IF
C
C  Check if an initial point is required.
C  In that case write the contents of array XP(1:N) to a file.
      IF (ICODE(1).EQ.1) THEN
         WRITE (IU,130) 'NOEVAL'
         DO 1,I=1,N
            CALL I2STR(I,ICHA,LECHA)
            WRITE (IU,110) 'POINT', ICHA(1:LECHA) , XP(I)
1        CONTINUE
         WRITE (IU,130) 'EVALUATE'
      END IF
C
C  Check if the fix-status  is required.
C  In that case write the contents of array IXAT(1:N) to a file.
      IF (ICODE(4).EQ.1) THEN
         DO 4,I=1,N
            IF (IXAT(I).EQ.0) THEN
               CALL I2STR(I,ICHA,LECHA)
               WRITE (IU,120) 'FIX', ICHA(1:LECHA)
            END IF
4        CONTINUE
      END IF
      WRITE (IU,130) 'FULLPRINT'
C
C  If the input is an MCL program, run it.
      IF (MCLIN) THEN
         WRITE (IU,120) 'RUNMCL',FINP(1:LEFIN)
         WRITE (IU,130) 'RETURN'
      END IF
      CLOSE (IU)
C
C  Check if a configuration file exists.
      INQUIRE (FILE=CFILE,EXIST=CFGEXI,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         WRITE (*,*) 'OPTIMA: Cannot inquire the configuration ',
     &               'file ', CFILE
         WRITE (*,*) '        Error code: ',IOE
         STOP
      END IF
C
C  Preserve the configuration file if necessary.
      OPEN (IU,FILE=CFILE)
      IF (CFGEXI) THEN
         OPEN (IUC,FILE=CFTMP)
10       READ (IU,'(A)',END=20) LINE
         LE = LENGTH(LINE)
         IF (LE.NE.0) THEN
            WRITE (IUC,'(A)') LINE(1:LE)
         ELSE
            WRITE (IUC,*)
         END IF
         GOTO 10
20       CONTINUE
         CLOSE (IUC)
      END IF
      CALL POSIT(IU)
C
C  Append our stuff in the Merlin configuration file.
      IF (MCLIN) THEN
C  The input file is an MCL program. Use the temporary macro
C  file an the Merlin input.
         WRITE (IU,*) 'INPUT_FILE ', MACRO
         IF (LEFOUT.NE.0)
     &       WRITE (IU,*) 'OUTPUT_FILE ', FOUT(1:LEFOUT)
      ELSE
C  The input file is (hopefuly) a sequence of Merlin commands.
C  Call the macro file only if any of the initializations is desired.
         WRITE (IU,*) 'INPUT_FILE ', FINP(1:LEFIN)
         IF (LEFOUT.NE.0)
     &       WRITE (IU,*) 'OUTPUT_FILE ', FOUT(1:LEFOUT)
         IF (ICODE(1).EQ.1 .OR. ICODE(2).EQ.1 .OR.
     &       ICODE(3).EQ.1 .OR. ICODE(4).EQ.1)
     &          WRITE (IU,*) 'PROLOG .'//MACRO
      END IF
      CLOSE (IU)
C
C  Call the minimizer.
      CALL MERLIN(N,M,VERSIM,MAXW,IQUIT)
C
C  Allocate a unit number for the deletion of temporary files.
      IU = NUNIT()
      IF (IU.EQ.-1) THEN
         WRITE (*,*) 'OPTIMA: Cannot allocate a free unit.'
         STOP
      END IF
C
C  Delete the temporary macro file.
      OPEN (IU,FILE=MACRO)
      CLOSE (IU,STATUS='DELETE')
C
      IF (CFGEXI) THEN
C  A Merlin configuration file existed before the routine was called.
C  Restore its contents.
         IUC = NUNIT()
         IF (IUC.EQ.-1) THEN
            WRITE (*,*) 'OPTIMA: Cannot allocate a free unit.'
            STOP
         END IF
         OPEN (IU,FILE=CFTMP)
         OPEN (IUC,FILE=CFILE)
30       READ (IU,'(A)',END=40) LINE
         LE = LENGTH(LINE)
         IF (LE.NE.0) THEN
            WRITE (IUC,'(A)') LINE(1:LE)
         ELSE
            WRITE (IUC,*)
         END IF
         GOTO 30
40       CONTINUE
         CLOSE (IUC)
         CLOSE (IU,STATUS='DELETE')
      ELSE
C  A Merlin configuration file did not exist before this routine
C  was called.
         OPEN (IU,FILE=CFILE)
         CLOSE (IU,STATUS='DELETE')
      END IF
C
C  Get the current point and store it to array XP(1:N)
      CALL GETX(XP)
C
C  Get the current value and store it to variable XV. ( XV=acsq(XP,N) )
      CALL GETVAL(XV)
C
C  Get the gradient vector and calculate the RMS gradient.
      CALL GETG(GRAD)
      GS = 0.
      DO 50,I=1,N
         GS = GS+GRAD(I)**2
50    CONTINUE
      GRMS = SQRT(GS/N)
C
C  Get the Merlin counters.
      CALL GETCNT(NF,NFP,NG,NGP,NH,NHP,NJ,NJP)
      RETURN
C
C  EOF encountered when reading the input file. Since we have already
C  checked that the input file exists, this situation means that the
C  input file is empty.
100   CONTINUE
      WRITE (*,*) 'OPTIMA: The input file ', FINP(1:LEFIN),
     &            ' is empty.'
      STOP
C
110   FORMAT (A,1X,A,1X,G21.14)
120   FORMAT (A,1X,A)
130   FORMAT (A)
      END
